from setuptools import setup, find_packages

setup(
    name='mcmc_diagnostics',
    version='0.1',
    description='Python analogue of the R coda package',
    url='https://github.com/aki-nishimura/mcmc-diagnostics',
    author='Akihiko (Aki) Nishimura',
    author_email='akihiko4@ucla.edu',
    license='MIT',
    packages=find_packages()
)